import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import (
    Description,
    FlavorText,
    GenerationGameIndex,
    Name,
    NamedResource,
    Resource,
    VerboseEffect,
    VersionEncounterDetail,
    VersionGameIndex,
)

from .utils import (
    AbilityEffectChange,
    AbilityFlavorText,
    AbilityPokemon,
    AwesomeName,
    Genus,
    GrowthRateExperienceLevel,
    MoveBattleStylePreference,
    MoveStatAffectSets,
    NaturePokeathlonStatAffectSet,
    NatureStatAffectSets,
    NatureStatChange,
    PalParkEncounterArea,
    PokemonAbility,
    PokemonFormSprites,
    PokemonHeldItem,
    PokemonMove,
    PokemonSpeciesDexEntry,
    PokemonSpeciesGender,
    PokemonSpeciesVariety,
    PokemonSprite,
    PokemonStat,
    PokemonType,
    PokemonTypePast,
    TypePokemon,
    TypeRelations,
)

__all__: t.Tuple[str, ...] = (
    "Ability",
    "Characteristic",
    "Pokemon",
    "PokemonSpecies",
    "PokemonForm",
    "Nature",
    "Type",
    "Gender",
    "GrowthRate",
    "EggGroup",
    "LocationAreaEncounter",
    "PokemonColor",
    "PokeathlonStat",
    "PokemonHabitats",
    "PokemonShape",
    "Stat",
)


@attrs.define(slots=True, kw_only=True)
class Ability(BaseModel):
    """Ability model.

    Attributes
    ----------
    id: int
        The identifier for this ability resource.
    name: str
        The name for this ability resource.
    is_main_series: bool
        Whether or not this ability originated in the main series of the video games.
    generation: NamedResource
        The generation this ability originated in.
    names: t.List[Name]
        The name of this ability listed in different languages.
    effect_entries: t.List[VerboseEffect]
        The effect of this ability listed in different languages.
    effect_changes: t.List[AbilityEffectChange]
        The list of previous effects this ability has had across version groups of the games.
    flavor_text_entries: t.List[AbilityFlavorText]
        The flavor text of this ability listed in different languages.
    pokemon: t.List[AbilityPokemon]
        A list of Pokémon that could potentially have this ability.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    is_main_series: bool = attrs.field(factory=bool)
    generation: NamedResource = attrs.field(factory=NamedResource)
    names: t.List[Name] = attrs.field(factory=list)
    effect_entries: t.List[VerboseEffect] = attrs.field(factory=list)
    effect_changes: t.List[AbilityEffectChange] = attrs.field(factory=list)
    flavor_text_entries: t.List[AbilityFlavorText] = attrs.field(factory=list)
    pokemon: t.List[AbilityPokemon] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Ability":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            is_main_series=payload.get("is_main_series", False),
            generation=NamedResource.from_payload(payload.get("generation", {})),
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            effect_entries=[VerboseEffect.from_payload(i) for i in payload.get("effect_entries", [])],
            effect_changes=[AbilityEffectChange.from_payload(i) for i in payload.get("effect_changes", [])],
            flavor_text_entries=[AbilityFlavorText.from_payload(i) for i in payload.get("flavor_text_entries", [])],
            pokemon=[AbilityPokemon.from_payload(i) for i in payload.get("pokemon", [])],
        )


@attrs.define(slots=True, kw_only=True)
class Characteristic(BaseModel):
    """Characteristic model.

    Attributes
    ----------
    id: int
        The identifier for this characteristic resource.
    gene_modulo: int
        The remainder of the highest stat/IV divided by 5.
    possible_values: t.List[int]
        The possible values of the highest stat that would result in a Pokémon
         recieving this characteristic when divided by 5.

    """

    id: int = attrs.field(factory=int)
    gene_modulo: int = attrs.field(factory=int)
    possible_values: t.List[int] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Characteristic":
        return cls(
            id=payload.get("id", 0),
            gene_modulo=payload.get("gene_modulo", 0),
            possible_values=payload.get("possible_values", []),
        )


@attrs.define(slots=True, kw_only=True)
class EggGroup(BaseModel):
    """EggGroup model.

    Attributes
    ----------
    id: int
        The identifier for this egg group resource.
    name: str
        The name for this egg group resource.
    names: t.List[Name]
        The name of this egg group listed in different languages.
    pokemon_species: t.List[NamedResource]
        A list of all Pokémon species that are members of this egg group.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    names: t.List[Name] = attrs.field(factory=list)
    pokemon_species: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "EggGroup":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            pokemon_species=[NamedResource.from_payload(i) for i in payload.get("pokemon_species", [])],
        )


@attrs.define(slots=True, kw_only=True)
class Gender(BaseModel):
    """Gender model.

    Attributes
    ----------
    id: int
        The identifier for gender resource.
    name: str
        The name for this gender resource.
    pokemon_species_details: t.List[PokemonSpeciesGender]
        A list of Pokémon species that belong to this gender.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    pokemon_species_details: t.List[PokemonSpeciesGender] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Gender":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            pokemon_species_details=[
                PokemonSpeciesGender.from_payload(i) for i in payload.get("pokemon_species_details", [])
            ],
        )


@attrs.define(slots=True, kw_only=True)
class GrowthRate(BaseModel):
    """GrowthRate model.

    Attributes
    ----------
    id: int
        The identifier for this growth rate resource.
    name: str
        The name for this growth rate resource.
    formula: str
        The formula used to calculate the rate at which the Pokémon species gains level.
    descriptions: t.List[Description]
        The descriptions of this characteristic listed in different languages.
    levels: t.List[GrowthRateExperienceLevel]
        A list of levels and the amount of experienced needed to atain them based on this growth rate.
    pokemon_species: t.List[NamedResource]
        A list of Pokémon species that gain levels at this growth rate.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    formula: str = attrs.field(factory=str)
    descriptions: t.List[Description] = attrs.field(factory=list)
    levels: t.List[GrowthRateExperienceLevel] = attrs.field(factory=list)
    pokemon_species: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "GrowthRate":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            formula=payload.get("formula", ""),
            descriptions=[Description.from_payload(i) for i in payload.get("descriptions", [])],
            levels=[GrowthRateExperienceLevel.from_payload(i) for i in payload.get("levels", [])],
            pokemon_species=[NamedResource.from_payload(i) for i in payload.get("pokemon_species", [])],
        )


@attrs.define(slots=True, kw_only=True)
class Nature(BaseModel):
    """Nature model.

    Attributes
    ----------
    id: int
        The identifier for this nature resource.
    name: str
        The name for this nature resource.
    decreased_stat: NamedAPIResource
        The stat decreased by 10% in Pokémon with this nature.
    increased_stat: NamedAPIResource
        The stat increased by 10% in Pokémon with this nature.
    hates_flavor: NamedAPIResource
        The flavor hated by Pokémon with this nature.
    likes_flavor: NamedAPIResource
        The flavor liked by Pokémon with this nature.
    pokeathlon_stat_changes: t.List[NatureStatChange]
        A list of Pokéathlon stats this nature effects and how much it effects them.
    move_battle_style_preferences: t.List[MoveBattleStylePreference]
        A list of battle styles and how likely a Pokémon with this nature
        is to use them in the Battle Palace or Battle Tent.
    names: t.List[Name]
        The name of this nature listed in different languages.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    decreased_stat: NamedResource = attrs.field(factory=NamedResource)
    increased_stat: NamedResource = attrs.field(factory=NamedResource)
    hates_flavor: NamedResource = attrs.field(factory=NamedResource)
    likes_flavor: NamedResource = attrs.field(factory=NamedResource)
    pokeathlon_stat_changes: t.List[NatureStatChange] = attrs.field(factory=list)
    move_battle_style_preferences: t.List[MoveBattleStylePreference] = attrs.field(factory=list)
    names: t.List[Name] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Nature":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            decreased_stat=NamedResource.from_payload(payload.get("decreased_stat", {})),
            increased_stat=NamedResource.from_payload(payload.get("increased_stat", {})),
            hates_flavor=NamedResource.from_payload(payload.get("hates_flavor", {})),
            likes_flavor=NamedResource.from_payload(payload.get("likes_flavor", {})),
            pokeathlon_stat_changes=[
                NatureStatChange.from_payload(i) for i in payload.get("pokeathlon_stat_changes", [])
            ],
            move_battle_style_preferences=[
                MoveBattleStylePreference.from_payload(i) for i in payload.get("move_battle_style_preferences", [])
            ],
            names=[Name.from_payload(i) for i in payload.get("names", [])],
        )


@attrs.define(slots=True, kw_only=True)
class PokeathlonStat(BaseModel):
    """PokeathlonStat model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    names: t.List[Name]
        A list of natures which affect this Pokéathlon stat positively or negatively.
    affecting_natures: NaturePokeathlonStatAffectSets
        A detail of natures which affect this Pokéathlon stat positively or negatively.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    names: t.List[Name] = attrs.field(factory=list)
    affecting_natures: NaturePokeathlonStatAffectSet = attrs.field(factory=NaturePokeathlonStatAffectSet)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokeathlonStat":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            affecting_natures=NaturePokeathlonStatAffectSet.from_payload(payload.get("affecting_natures", {})),
        )


@attrs.define(slots=True, kw_only=True)
class Pokemon(BaseModel):
    """Pokemon model.

    Attributes
    ----------
    id: int
        The identifier for this Pokémon resource.
    name: str
        The name for this Pokémon resource.
    base_experience: int
        The base experience gained for defeating this Pokémon.
    height: int
        The height of this Pokémon in decimetres.
    is_default: bool
        Set for exactly one Pokémon used as the default for each species.
    order: int
        Order for sorting. Almost national order, except families are grouped together.
    weight: int
        The weight of this Pokémon in hectograms.
    abilities: t.List[PokemonAbility]
        A list of abilities this Pokémon could potentially have.
    forms: t.List[NamedAPIResource]
        A list of forms this Pokémon can take on.
    game_indices: t.List[VersionGameIndex]
        A list of game indices relevent to Pokémon item by generation.
    held_items: t.List[PokemonHeldItem]
        A list of items this Pokémon may be holding when encountered.
    location_area_encounters: str
        Location area encounter details for different versions.
    moves: t.List[PokemonMove]
        A list of details showing types this Pokémon has.
    species: NamedAPIResource
        The species this Pokémon belongs to.
    sprites: PokemonSprites
        A set of sprites used to depict this Pokémon in the game.
    stats: t.List[PokemonStat]
        A list of details showing all the stats this Pokémon has.
    types: t.List[PokemonType]
        A list of details showing types this Pokémon has.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    base_experience: int = attrs.field(factory=int)
    height: int = attrs.field(factory=int)
    is_default: bool = attrs.field(factory=bool)
    order: int = attrs.field(factory=int)
    weight: int = attrs.field(factory=int)
    abilities: t.List[PokemonAbility] = attrs.field(factory=list)
    forms: t.List[NamedResource] = attrs.field(factory=list)
    game_indices: t.List[VersionGameIndex] = attrs.field(factory=list)
    held_items: t.List[PokemonHeldItem] = attrs.field(factory=list)
    location_area_encounters: str = attrs.field(factory=str)
    moves: t.List[PokemonMove] = attrs.field(factory=list)
    past_types: t.List[PokemonTypePast] = attrs.field(factory=list)
    sprites: PokemonSprite = attrs.field(factory=PokemonSprite)
    species: NamedResource = attrs.field(factory=NamedResource)
    stats: t.List[PokemonStat] = attrs.field(factory=list)
    types: t.List[PokemonType] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Pokemon":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            base_experience=payload.get("base_experience", 0),
            height=payload.get("height", 0),
            is_default=payload.get("is_default", False),
            order=payload.get("order", 0),
            weight=payload.get("weight", 0),
            abilities=[PokemonAbility.from_payload(i) for i in payload.get("abilities", [])],
            forms=[NamedResource.from_payload(i) for i in payload.get("forms", [])],
            game_indices=[VersionGameIndex.from_payload(i) for i in payload.get("game_indices", [])],
            held_items=[PokemonHeldItem.from_payload(i) for i in payload.get("held_items", [])],
            location_area_encounters=payload.get("location_area_encounters", ""),
            moves=[PokemonMove.from_payload(i) for i in payload.get("moves", [])],
            past_types=[PokemonTypePast.from_payload(i) for i in payload.get("past_types", [])],
            sprites=PokemonSprite.from_payload(payload.get("sprites", {})),
            species=NamedResource.from_payload(payload.get("species", {})),
            stats=[PokemonStat.from_payload(i) for i in payload.get("stats", [])],
            types=[PokemonType.from_payload(i) for i in payload.get("types", [])],
        )


@attrs.define(slots=True, kw_only=True)
class LocationAreaEncounter(BaseModel):
    """LocationAreaEncounter model.

    Attributes
    ----------
    location_area: NamedAPIResource
        The location area the referenced Pokémon can be encountered in.
    version_details: t.List[VersionEncounterDetail]
        A list of versions and encounters with the referenced Pokémon that might happen.
    """

    location_area: NamedResource = attrs.field(factory=NamedResource)
    version_details: t.List[VersionEncounterDetail] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "LocationAreaEncounter":
        return cls(
            location_area=NamedResource.from_payload(payload.get("location_area", {})),
            version_details=[VersionEncounterDetail.from_payload(i) for i in payload.get("version_details", [])],
        )


@attrs.define(slots=True, kw_only=True)
class PokemonColor(BaseModel):
    """PokemonColor model.

    Attributes
    ----------
    id: int
        The identifier for this Pokémon color resource.
    name: str
        The name for this Pokémon color resource.
    names: t.List[Name]
        The name of this Pokémon color listed in different languages.
    pokemon_species: t.List[NamedAPIResource]
        A list of the Pokémon species that have this color.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    names: t.List[Name] = attrs.field(factory=list)
    pokemon_species: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonColor":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            pokemon_species=[NamedResource.from_payload(i) for i in payload.get("pokemon_species", [])],
        )


@attrs.define(slots=True, kw_only=True)
class PokemonForm(BaseModel):
    """PokemonForm model.

    Attributes
    ----------
    id: int
        The identifier for this Pokémon form resource.
    name: str
        The name for this Pokémon form resource.
    order: int
        The order in which forms should be sorted within all forms. Multiple forms may have equal order,
        in which case they should fall back on sorting by name.
    form_order: int
        The order in which forms should be sorted within a species' forms.
    is_default: bool
        True for exactly one form used as the default for each Pokémon.
    is_battle_only: bool
        Whether or not this form can only happen during battle.
    is_mega: bool
        Whether or not this form requires mega evolution.
    form_name: str
        The name of this form.
    pokemon: NamedAPIResource
        The Pokémon that can take on this form.
    sprites: PokemonFormSprites
        A set of sprites used to depict this Pokémon form in the game.
    version_group: NamedAPIResource
        The version group this Pokémon form was introduced in.
    names: t.List[Name]
        The form specific full name of this Pokémon form, or empty if the form does not have a specific name.
    form_names: t.List[Name]
        The form specific form name of this Pokémon form, or empty if the form does not have a specific name.
    pokemon_species: NamedAPIResource
        The Pokémon species that this form belongs to.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    order: int = attrs.field(factory=int)
    form_order: int = attrs.field(factory=int)
    is_default: bool = attrs.field(factory=bool)
    is_battle_only: bool = attrs.field(factory=bool)
    is_mega: bool = attrs.field(factory=bool)
    form_name: str = attrs.field(factory=str)
    pokemon: NamedResource = attrs.field(factory=NamedResource)
    sprites: PokemonFormSprites = attrs.field(factory=PokemonFormSprites)
    version_group: NamedResource = attrs.field(factory=NamedResource)
    names: t.List[Name] = attrs.field(factory=list)
    form_names: t.List[Name] = attrs.field(factory=list)
    pokemon_species: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonForm":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            order=payload.get("order", 0),
            form_order=payload.get("form_order", 0),
            is_default=payload.get("is_default", False),
            is_battle_only=payload.get("is_battle_only", False),
            is_mega=payload.get("is_mega", False),
            form_name=payload.get("form_name", ""),
            pokemon=NamedResource.from_payload(payload.get("pokemon", {})),
            sprites=PokemonFormSprites.from_payload(payload.get("sprites", {})),
            version_group=NamedResource.from_payload(payload.get("version_group", {})),
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            form_names=[Name.from_payload(i) for i in payload.get("form_names", [])],
            pokemon_species=NamedResource.from_payload(payload.get("pokemon_species", {})),
        )


@attrs.define(slots=True, kw_only=True)
class PokemonHabitats(BaseModel):
    """PokemonHabitats model.

    Attributes
    ----------
    id: int
        The identifier for this Pokémon habitat resource.
    name: str
        The name for this Pokémon habitat resource.
    names: t.List[Name]
        The name of this Pokémon habitat listed in different languages.
    pokemon_species: t.List[NamedAPIResource]
        A list of the Pokémon species that can be found in this habitat.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    names: t.List[Name] = attrs.field(factory=list)
    pokemon_species: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonHabitats":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            pokemon_species=[NamedResource.from_payload(i) for i in payload.get("pokemon_species", [])],
        )


@attrs.define(slots=True, kw_only=True)
class PokemonShape(BaseModel):
    """PokemonShape model.

    Attributes
    ----------
    id: int
        The identifier for this Pokémon shape resource.
    name: str
        The name for this Pokémon shape resource.
    awesome_names: t.List[AwesomeName]
        The "scientific" name of this Pokémon shape listed in different languages.
    names: t.List[Name]
        The name of this Pokémon shape listed in different languages.
    pokemon_species: t.List[NamedAPIResource]
        A list of the Pokémon species that have this shape.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    awesome_names: t.List[AwesomeName] = attrs.field(factory=list)
    names: t.List[Name] = attrs.field(factory=list)
    pokemon_species: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonShape":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            awesome_names=[AwesomeName.from_payload(i) for i in payload.get("awesome_names", [])],
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            pokemon_species=[NamedResource.from_payload(i) for i in payload.get("pokemon_species", [])],
        )


@attrs.define(slots=True, kw_only=True)
class PokemonSpecies(BaseModel):
    """PokemonSpecies model.

    Attributes
    ----------
    id: int
        The identifier for this Pokémon species resource.
    name: str
        The name for this Pokémon species resource.
    order: int
        The order in which species should be sorted. Based on National Dex order,
        except families are grouped together and sorted by stage.
    gender_rate: int
        The chance of this Pokémon being of a particular gender. Ratio is male to female. -1 for genderless.
    capture_rate: int
        The base capture rate; up to 255. The higher the number, the easier the catch.
    base_happiness: int
        The happiness when caught by a normal Pokéball; up to 255. The higher the number, the happier the Pokémon.
    is_baby: bool
        Whether or not this is a baby Pokémon.
    is_legendary: bool
        Whether or not this is a legendary Pokémon.
    is_mythical: bool
        Whether or not this is a mythical Pokémon.
    hatch_counter: int
        Initial hatch counter: one must walk 255 × (hatch_counter + 1) steps before this
        Pokémon's egg hatches, unless utilizing bonuses like Flame Body's.
    has_gender_differences: bool
        Whether or not this Pokémon has visual differences due to gender.
    forms_switchable: bool
        Whether or not this Pokémon has multiple forms and can switch between them.
    growth_rate: NamedAPIResource
        The rate at which this Pokémon species gains levels.
    pokedex_numbers: t.List[PokemonSpeciesDexEntry]
        A list of Pokedexes and the indexes reserved within them for this Pokémon species.
    egg_groups: t.List[NamedAPIResource]
        A list of egg groups this Pokémon species is a member of.
    color: NamedAPIResource
        The color of this Pokémon for Pokédex search.
    shape: NamedAPIResource
        The shape of this Pokémon for Pokédex search.
    evolves_from_species: NamedAPIResource
        The Pokémon species that evolves into this Pokemon_species.
    evolution_chain: APIResource
        The evolution chain this Pokémon species is a member of.
    habitat: NamedAPIResource
        The habitat this Pokémon species can be encountered in.
    generation: NamedAPIResource
        The generation this Pokémon species was introduced in.
    names: t.List[Name]
        The name of this Pokémon species listed in different languages.
    pal_park_encounters: t.List[PalParkEncounterArea]
        A list of encounters that can be had with this Pokémon species in pal park.
    flavor_text_entries: t.List[FlavorText]
        A list of flavor text entries for this Pokémon species.
    form_descriptions: t.List[Description]
        A list of form description for this Pokémon species.
    genera: t.List[Genus]
        A list of the genus of this Pokémon species listed in multiple languages.
    varieties: t.List[PokemonSpeciesVariety]
        A list of the Pokémon that exist within this Pokémon species.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    order: int = attrs.field(factory=int)
    gender_rate: int = attrs.field(factory=int)
    capture_rate: int = attrs.field(factory=int)
    base_happiness: int = attrs.field(factory=int)
    is_baby: bool = attrs.field(factory=bool)
    is_legendary: bool = attrs.field(factory=bool)
    is_mythical: bool = attrs.field(factory=bool)
    hatch_counter: int = attrs.field(factory=int)
    has_gender_differences: bool = attrs.field(factory=bool)
    forms_switchable: bool = attrs.field(factory=bool)
    growth_rate: NamedResource = attrs.field(factory=NamedResource)
    pokedex_numbers: t.List[PokemonSpeciesDexEntry] = attrs.field(factory=list)
    egg_groups: t.List[NamedResource] = attrs.field(factory=list)
    color: NamedResource = attrs.field(factory=NamedResource)
    shape: NamedResource = attrs.field(factory=NamedResource)
    evolves_from_species: NamedResource = attrs.field(factory=NamedResource)
    evolution_chain: Resource = attrs.field(factory=Resource)
    habitat: NamedResource = attrs.field(factory=NamedResource)
    generation: NamedResource = attrs.field(factory=NamedResource)
    names: t.List[Name] = attrs.field(factory=list)
    pal_park_encounters: t.List[PalParkEncounterArea] = attrs.field(factory=list)
    flavor_text_entries: t.List[FlavorText] = attrs.field(factory=list)
    form_descriptions: t.List[Description] = attrs.field(factory=list)
    genera: t.List[Genus] = attrs.field(factory=list)
    varieties: t.List[PokemonSpeciesVariety] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonSpecies":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            order=payload.get("order", 0),
            gender_rate=payload.get("gender_rate", 0),
            capture_rate=payload.get("capture_rate", 0),
            base_happiness=payload.get("base_happiness", 0),
            is_baby=payload.get("is_baby", False),
            is_legendary=payload.get("is_legendary", False),
            is_mythical=payload.get("is_mythical", False),
            hatch_counter=payload.get("hatch_counter", 0),
            has_gender_differences=payload.get("has_gender_differences", False),
            forms_switchable=payload.get("forms_switchable", False),
            growth_rate=NamedResource.from_payload(payload.get("growth_rate", {})),
            pokedex_numbers=[PokemonSpeciesDexEntry.from_payload(i) for i in payload.get("pokedex_numbers", [])],
            egg_groups=[NamedResource.from_payload(i) for i in payload.get("egg_groups", [])],
            color=NamedResource.from_payload(payload.get("color", {})),
            shape=NamedResource.from_payload(payload.get("shape", {})),
            evolves_from_species=NamedResource.from_payload(payload.get("evolves_from_species", {}) or {}),
            evolution_chain=Resource.from_payload(payload.get("evolution_chain", {}) or {}),
            habitat=NamedResource.from_payload(payload.get("habitat", {}) or {}),
            generation=NamedResource.from_payload(payload.get("generation", {})),
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            pal_park_encounters=[PalParkEncounterArea.from_payload(i) for i in payload.get("pal_park_encounters", [])],
            flavor_text_entries=[FlavorText.from_payload(i) for i in payload.get("flavor_text_entries", [])],
            form_descriptions=[Description.from_payload(i) for i in payload.get("form_descriptions", [])],
            genera=[Genus.from_payload(i) for i in payload.get("genera", [])],
            varieties=[PokemonSpeciesVariety.from_payload(i) for i in payload.get("varieties", [])],
        )


@attrs.define(slots=True, kw_only=True)
class Stat(BaseModel):
    """
    A Pokémon stat model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    game_index: int
        The stat order in which effects of this stat take place during battle.
    is_battle_only: bool
        Whether this stat only exists within a battle.
    affecting_moves: MoveStatAffectSets
        A detail of moves which affect this stat positively or negatively.
    affecting_natures: NatureStatAffectSets
        A detail of natures which affect this stat positively or negatively.
    characteristics: t.List[Resource]
        A list of characteristics that are set on a Pokémon when its highest base stat is this stat.
    move_damage_class: NamedResource
        The class of damage this stat is directly related to.
    names: t.List[Name]
        The name of this resource listed in different languages.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    game_index: int = attrs.field(factory=int)
    is_battle_only: bool = attrs.field(factory=bool)
    affecting_moves: MoveStatAffectSets = attrs.field(factory=MoveStatAffectSets)
    affecting_natures: NatureStatAffectSets = attrs.field(factory=NatureStatAffectSets)
    characteristics: t.List[Resource] = attrs.field(factory=list)
    move_damage_class: NamedResource = attrs.field(factory=NamedResource)
    names: t.List[Name] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Stat":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            game_index=payload.get("game_index", 0),
            is_battle_only=payload.get("is_battle_only", False),
            affecting_moves=MoveStatAffectSets.from_payload(payload.get("affecting_moves", {})),
            affecting_natures=NatureStatAffectSets.from_payload(payload.get("affecting_natures", {})),
            characteristics=[Resource.from_payload(i) for i in payload.get("characteristics", [])],
            move_damage_class=NamedResource.from_payload(payload.get("move_damage_class", {}) or {}),
            names=[Name.from_payload(i) for i in payload.get("names", [])],
        )


@attrs.define(slots=True, kw_only=True)
class Type(BaseModel):
    """
    A Pokémon type model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    damage_relations: TypeRelations
        A detail of how effective this type is toward others and vice versa.
    game_indices: t.List[GenerationGameIndex]
        A list of game indices relevent to this item by generation.
    generation: NamedResource
        The generation this type was introduced in.
    move_damage_class: NamedResource
        The class of damage inflicted by this type.
    names: t.List[Name]
        The name of this resource listed in different languages.
    pokemon: t.List[TypePokemon]
        A list of details of Pokémon that have this type.
    moves: t.List[NamedResource]
        A list of moves that have this type.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    damage_relations: TypeRelations = attrs.field(factory=TypeRelations)
    game_indices: t.List[GenerationGameIndex] = attrs.field(factory=list)
    generation: NamedResource = attrs.field(factory=NamedResource)
    move_damage_class: NamedResource = attrs.field(factory=NamedResource)
    names: t.List[Name] = attrs.field(factory=list)
    pokemon: t.List[TypePokemon] = attrs.field(factory=list)
    moves: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Type":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            damage_relations=TypeRelations.from_payload(payload.get("damage_relations", {})),
            game_indices=[GenerationGameIndex.from_payload(i) for i in payload.get("game_indices", [])],
            generation=NamedResource.from_payload(payload.get("generation", {})),
            move_damage_class=NamedResource.from_payload(payload.get("move_damage_class", {})),
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            pokemon=[TypePokemon.from_payload(i) for i in payload.get("pokemon", [])],
            moves=[NamedResource.from_payload(i) for i in payload.get("moves", [])],
        )



import asyncio
import time
import typing as t

import aiohttp

from pokelance.cache import Cache
from pokelance.exceptions import HTTPException

from .endpoints import Endpoint, Route

if t.TYPE_CHECKING:
    from pokelance import PokeLance


__all__: t.Tuple[str, ...] = (
    "HttpClient",
    "Route",
    "Endpoint",
)


@t.final
class HttpClient:
    """The HTTP client for PokeLance.

    Parameters
    ----------
    client: pokelance.PokeLance
        The client that this HTTP client is for.
    cache_size: int
        The size of the cache.
    session: aiohttp.ClientSession
        The session to use for the HTTP client.

    Attributes
    ----------
    session: aiohttp.ClientSession
        The session to use for the HTTP client.
    _is_ready: bool
        Whether the HTTP client is ready.
    _cache: pokelance.cache.Cache
        The cache to use for the HTTP client.
    _client: pokelance.PokeLance
        The client that this HTTP client is for.
    """

    __slots__: t.Tuple[str, ...] = (
        "_client",
        "session",
        "_cache",
        "_is_ready",
    )

    def __init__(
        self, *, cache_size: int, client: "PokeLance", session: t.Optional[aiohttp.ClientSession] = None
    ) -> None:
        """Initializes the HTTP client.

        Parameters
        ----------
        cache_size: int
            The size of the cache.
        client: pokelance.PokeLance
            The client that this HTTP client is for.
        session: aiohttp.ClientSession
            The session to use for the HTTP client.

        Returns
        -------
        pokelance.http.HttpClient
            The HTTP client.
        """
        self._client = client
        self.session = session
        self._is_ready = False
        self._cache = Cache(max_size=cache_size)

    def _load_endpoints(self) -> None:
        """Loads the endpoints for the HTTP client."""
        for num, (coro, name) in enumerate(self._client.loaders):
            task = asyncio.create_task(coro)
            task.add_done_callback(
                lambda _: self._client.logger.info(f"Loaded {self._client.loaders.pop(0)[1]} endpoints.")
            )
        self._is_ready = True

    async def connect(self) -> None:
        """Connects the HTTP client."""
        if not self._is_ready:
            if self.session is None:
                self.session = aiohttp.ClientSession()
                self._load_endpoints()

    async def request(self, route: Route) -> t.Any:
        """Makes a request to the PokeAPI.

        Parameters
        ----------
        route: pokelance.http.Route
            The route to use for the request.

        Returns
        -------
        t.Any
            The response from the PokeAPI.

        Raises
        ------
        pokelance.exceptions.HTTPException
            An error occurred while making the request.
        """
        if self.session is None:
            await self.connect()
        if not self._is_ready and self._client.loaders:
            self._load_endpoints()
        if self.session is not None:
            async with self.session.request(route.method, route.url, params=route.payload) as response:
                if 300 > response.status >= 200:
                    return await response.json()
                else:
                    raise HTTPException(str(response.reason), route, response.status).create()
        else:
            raise HTTPException("No session was provided.", route, 0).create()

    async def ping(self) -> float:
        """Pings the PokeAPI and returns the latency.

        Returns
        -------
        float
            The latency of the PokeAPI.
        """
        start = time.perf_counter()
        await self.request(Route())
        return time.perf_counter() - start

    @property
    def cache(self) -> Cache:
        """The cache to use for the HTTP client.

        Returns
        -------
        pokelance.cache.Cache
            The cache.
        """
        return self._cache



import typing as t

if t.TYPE_CHECKING:
    from pokelance.http import Route


__all__: t.Tuple[str, ...] = (
    "PokeLanceException",
    "HTTPException",
    "ResourceNotFound",
)


class PokeLanceException(Exception):
    """Base exception class for PokeLance.

    Parameters
    ----------
    message: str
        The message to display.
    route: Optional[pokelance.http.Route]
        The route that caused the exception.

    Attributes
    ----------
    message: str
        The message to display.
    route: pokelance.http.Route
        The route that caused the exception.

    """

    def __init__(self, message: str, route: "Route") -> None:
        self.message = message
        self.route = route
        super().__init__(message)

    def __str__(self) -> str:
        return f"{self.message} | {self.route}"


class HTTPException(PokeLanceException):
    """Base exception class for HTTP exceptions.

    Parameters
    ----------
    message: str
        The message to display.
    route: pokelance.http.Route
        The route that caused the exception.
    status: int
        The status code of the exception.

    Attributes
    ----------
    message: str
        The message to display.
    route: pokelance.http.Route
        The route that caused the exception.
    status: int
        The status code of the exception.
    """

    def __init__(self, message: str, route: "Route", status: int) -> None:
        self.status = status
        super().__init__(message, route)

    def __str__(self) -> str:
        return f"{self.message} | {self.route} | {self.status}"

    def create(self) -> "HTTPException":
        """Creates an exception from the status code."""
        return get_exception(self.status)(self.message, self.route, self.status)


class ResourceNotFound(PokeLanceException):
    """Exception raised when a resource is not found.

    Parameters
    ----------
    message: str
        The message to display.
    route: pokelance.http.Route
        The route that caused the exception.
    """

    def __init__(self, message: str, route: "Route") -> None:
        super().__init__(message, route)

    def __str__(self) -> str:
        return f"{self.message} | {self.route}"


class BadRequest(HTTPException):
    """Exception raised when a bad request is made. [HTTP 400]"""


class Unauthorized(HTTPException):
    """Exception raised when unauthorized. [HTTP 401]"""


class Forbidden(HTTPException):
    """Exception raised when forbidden. [HTTP 403]"""


class NotFound(HTTPException):
    """Exception raised when a resource is not found. [HTTP 404]"""


class MethodNotAllowed(HTTPException):
    """Exception raised when a method is not allowed. [HTTP 405]"""


class UnknownError(HTTPException):
    """Exception raised when an unknown error occurs."""


CODES: t.Dict[int, t.Type[HTTPException]] = {
    400: BadRequest,
    401: Unauthorized,
    403: Forbidden,
    404: NotFound,
    405: MethodNotAllowed,
}


def get_exception(status: int) -> t.Type[HTTPException]:
    """Gets an exception from the status code.

    Parameters
    ----------
    status: int
        The status code.

    Returns
    -------
    pokelance.exceptions.HTTPException
        The exception.
    """
    return CODES.get(status, UnknownError)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pokelance',
 'pokelance.cache',
 'pokelance.ext',
 'pokelance.http',
 'pokelance.models',
 'pokelance.models.abstract',
 'pokelance.models.abstract.utils',
 'pokelance.models.common']

package_data = \
{'': ['*']}

install_requires = \
['aiofiles>=22.1.0,<23.0.0',
 'aiohttp>=3.8.3,<4.0.0',
 'attrs>=22.1.0,<23.0.0',
 'types-aiofiles>=22.1.0.4,<23.0.0.0']

setup_kwargs = {
    'name': 'pokelance',
    'version': '0.0.1a0',
    'description': 'A flexible and easy to use pokemon library.',
    'long_description': '<h1 align="center"><b>PokeLance</b></h1>\n<p align="center">\n<img src="docs/assets/pokelance.png" width=450 alt="logo"><br><br>\n<img src="https://img.shields.io/github/license/FallenDeity/PokeLance?style=flat-square" alt="license">\n<img src="https://img.shields.io/badge/code%20style-black-000000.svg?style=flat-square" alt="black">\n<img src="https://img.shields.io/badge/%20type_checker-mypy-%231674b1?style=flat-square" alt="mypy">\n<img src="https://img.shields.io/badge/%20linter-ruff-%231674b1?style=flat-square" alt="ruff">\n<img src="https://img.shields.io/github/stars/FallenDeity/PokeLance?style=flat-square" alt="stars">\n<img src="https://img.shields.io/github/last-commit/FallenDeity/PokeLance?style=flat-square" alt="commits">\n<img src="https://img.shields.io/pypi/pyversions/PokeLance?style=flat-square" alt="py">\n<img src="https://img.shields.io/pypi/v/PokeLance?style=flat-square" alt="versions">\n<br><br>\nA flexible, statically typed and easy to use pokeapi wrapper for python 🚀\n</p>\n\n---\n\n\nFeatures:\n- Modern and pythonic API asynchronously built on top of aiohttp\n- Flexible and easy to use\n- Statically typed with mypy\n- Linted with ruff\n- Well documented\n- Optimized for speed and performance\n- Automatically caches data for faster access\n- Caches endpoints for user convenience\n\n---\n\n## Installation\n\n```bash\n$ python -m pip install PokeLance\n```\n\n---\n\n## Usage\n\n```python\nimport asyncio\n\nfrom pokelance import PokeLance\n\nclient = PokeLance()  # Create a client instance\n\n\nasync def main() -> None:\n    print(await client.ping())  # Ping the pokeapi\n    print(await client.berry.fetch_berry("cheri"))  # Fetch a berry from the pokeapi\n    print(await client.berry.fetch_berry_flavor("spicy"))\n    print(await client.berry.fetch_berry_firmness("very-soft"))\n    print(client.berry.get_berry("cheri"))  # Get a cached berry it will return None if it doesn\'t exist\n    print(client.berry.get_berry_flavor("spicy"))\n    print(client.berry.get_berry_firmness("very-soft"))\n    await client.close()  # Close the client\n    return None\n\n\nasyncio.run(main())\n```\n\n## With Async Context Manager\n\n```python\nimport asyncio\n\nimport aiohttp\nfrom pokelance import PokeLance\n\n\nasync def main() -> None:\n    # Use an async context manager to create a client instance\n    async with aiohttp.ClientSession() as session, PokeLance(session=session) as client:\n        print(await client.ping())  # Ping the pokeapi\n        print(await client.berry.fetch_berry("cheri"))  # Fetch a berry from the pokeapi\n        print(await client.berry.fetch_berry_flavor("spicy"))\n        print(await client.berry.fetch_berry_firmness("very-soft"))\n        print(client.berry.get_berry("cheri"))  # Get a cached berry it will return None if it doesn\'t exist\n        print(client.berry.get_berry_flavor("spicy"))\n        print(client.berry.get_berry_firmness("very-soft"))\n        # The client will be closed automatically when the async context manager exits\n    return None\n\nasyncio.run(main())\n```\n\n## Important Links\n\n\n- [PokeAPI](https://pokeapi.co/)\n- [PokeAPI Documentation](https://pokeapi.co/docs/v2)\n- [PokeLance Documentation](https://FallenDeity.github.io/PokeLance/)\n- [PokeLance ReadTheDocs](https://pokelance.readthedocs.io/en/latest/)\n- [PokeLance GitHub](https://github.com/FallenDeity/PokeLance)\n- [PokeLance PyPI](https://pypi.org/project/PokeLance/)\n- [PokeLance Discord](https://discord.gg/yeyEvT5V2J)\n\n---\n\n!!! note "Note"\n    This is a work in progress. If you find any bugs or have any suggestions, please open an issue [here](https://github.com/FallenDeity/PokeLance/issues/new).\n\n---',
    'author': 'FallenDeity',
    'author_email': '61227305+FallenDeity@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import typing as t

from .berries import BerryFlavorMap, FlavorBerryMap
from .contests import ContestName
from .evolutions import ChainLink, EvolutionDetail
from .games import PokemonEntry
from .items import ItemHolderPokemon, ItemHolderPokemonVersionDetail, ItemSprites
from .locations import EncounterMethodRate, EncounterVersionDetails, PalParkEncounterSpecies, PokemonEncounter
from .moves import ContestComboDetail, ContestComboSet, MoveFlavorText, MoveMetaData, MoveStatChange, PastMoveStatValues
from .pokemons import (
    AbilityEffectChange,
    AbilityFlavorText,
    AbilityPokemon,
    AwesomeName,
    Genus,
    GrowthRateExperienceLevel,
    MoveBattleStylePreference,
    MoveStatAffectSets,
    MoveStatEffect,
    NaturePokeathlonStatAffect,
    NaturePokeathlonStatAffectSet,
    NatureStatAffectSets,
    NatureStatChange,
    PalParkEncounterArea,
    PokemonAbility,
    PokemonFormSprites,
    PokemonFormType,
    PokemonHeldItem,
    PokemonHeldItemVersion,
    PokemonMove,
    PokemonMoveVersion,
    PokemonSpeciesDexEntry,
    PokemonSpeciesGender,
    PokemonSpeciesVariety,
    PokemonSprite,
    PokemonStat,
    PokemonType,
    PokemonTypePast,
    TypePokemon,
    TypeRelations,
    TypeRelationsPast,
)

__all__: t.Tuple[str, ...] = (
    "BerryFlavorMap",
    "FlavorBerryMap",
    "ContestName",
    "EvolutionDetail",
    "ChainLink",
    "PokemonEntry",
    "ItemHolderPokemonVersionDetail",
    "ItemHolderPokemon",
    "ItemSprites",
    "EncounterVersionDetails",
    "EncounterMethodRate",
    "PokemonEncounter",
    "PalParkEncounterSpecies",
    "ContestComboDetail",
    "ContestComboSet",
    "MoveFlavorText",
    "MoveMetaData",
    "MoveStatChange",
    "PastMoveStatValues",
    "AbilityEffectChange",
    "AbilityPokemon",
    "AbilityFlavorText",
    "PokemonSpeciesGender",
    "GrowthRateExperienceLevel",
    "NatureStatChange",
    "MoveBattleStylePreference",
    "NaturePokeathlonStatAffect",
    "NaturePokeathlonStatAffectSet",
    "PokemonAbility",
    "PokemonFormType",
    "PokemonHeldItemVersion",
    "PokemonHeldItem",
    "PokemonMoveVersion",
    "PokemonMove",
    "PokemonSprite",
    "PokemonStat",
    "PokemonTypePast",
    "PokemonType",
    "PokemonFormSprites",
    "AwesomeName",
    "Genus",
    "PokemonSpeciesDexEntry",
    "PokemonSpeciesVariety",
    "PalParkEncounterArea",
    "NatureStatAffectSets",
    "MoveStatEffect",
    "MoveStatAffectSets",
    "TypePokemon",
    "TypeRelationsPast",
    "TypeRelations",
)



import typing as t

from pokelance.http import Endpoint
from pokelance.models import Machine as MachineModel

from ._base import BaseExtension

if t.TYPE_CHECKING:
    from pokelance import PokeLance
    from pokelance.http import HttpClient


__all__: t.Tuple[str, ...] = (
    "setup",
    "Machine",
)


class Machine(BaseExtension):
    """Extension for machine related endpoints.

    Parameters
    ----------
    client: pokelance.http.HttpClient
        The client to use for requests.

    Attributes
    ----------
    cache: pokelance.cache.Machine
        The cache for this extension.
    """

    def __init__(self, client: "HttpClient") -> None:
        """Initializes the extension."""
        super().__init__(client)
        self.cache = self._cache.machine

    async def setup(self) -> None:
        """Sets up the extension."""
        data = await self._client.request(Endpoint.get_machine_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "machine", data["results"])

    def get_machine(self, id_: int) -> t.Optional[MachineModel]:
        """Gets a machine from the cache.

        Parameters
        ----------
        id_: int
            The id of the machine.

        Returns
        -------
        typing.Optional[pokelance.models.Machine]
            The machine if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the machine does not exist in the cache.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> machine = client.machine.get_machine(1)
        >>> machine.item.name
        'tm00'
        """
        route = Endpoint.get_machine(id_)
        self._validate_resource(self.cache.machine, id_, route)
        return self.cache.machine.get(route, None)

    async def fetch_machine(self, id_: int) -> t.Optional[MachineModel]:
        """Fetches a machine from the API.

        Parameters
        ----------
        id_: int
            The id of the machine.

        Returns
        -------
        typing.Optional[pokelance.models.Machine]
            The machine if it exists in the API, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the machine does not exist in the API.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     machine = await client.machine.fetch_machine(1)
        ...     print(machine.item.name)
        ...     await client.close()
        >>> asyncio.run(main())
        tm00
        """
        route = Endpoint.get_machine(id_)
        self._validate_resource(self.cache.machine, id_, route)
        data = await self._client.request(route)
        return self.cache.machine.setdefault(route, MachineModel.from_payload(data))


def setup(lance: "PokeLance") -> None:
    """Sets up the machine cog."""
    lance.add_extension("machine", Machine(lance.http))



import typing as t

from pokelance.http import Endpoint
from pokelance.models import (
    Ability,
    Characteristic,
    EggGroup,
    Gender,
    GrowthRate,
    LocationAreaEncounter,
    Nature,
    PokeathlonStat,
)
from pokelance.models import Pokemon as PokemonModel
from pokelance.models import PokemonColor, PokemonForm, PokemonHabitats, PokemonShape, PokemonSpecies, Stat, Type

from ._base import BaseExtension

if t.TYPE_CHECKING:
    from pokelance import PokeLance
    from pokelance.http import HttpClient


__all__: t.Tuple[str, ...] = (
    "setup",
    "Pokemon",
)


class Pokemon(BaseExtension):
    """Extension for pokemon related endpoints.

    Parameters
    ----------
    client: pokelance.http.HttpClient
        The client to use for requests.

    Attributes
    ----------
    cache: pokelance.cache.Pokemon
        The cache for this extension.
    """

    def __init__(self, client: "HttpClient") -> None:
        """Initialize the extension."""
        super().__init__(client)
        self.cache = self._cache.pokemon

    async def setup(self) -> None:
        """Sets up the extension."""
        data = await self._client.request(Endpoint.get_ability_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "ability", data["results"])
        data = await self._client.request(Endpoint.get_characteristic_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "characteristic", data["results"])
        data = await self._client.request(Endpoint.get_egg_group_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "egg_group", data["results"])
        data = await self._client.request(Endpoint.get_gender_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "gender", data["results"])
        data = await self._client.request(Endpoint.get_growth_rate_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "growth_rate", data["results"])
        data = await self._client.request(Endpoint.get_nature_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "nature", data["results"])
        data = await self._client.request(Endpoint.get_pokeathlon_stat_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "pokeathlon_stat", data["results"])
        data = await self._client.request(Endpoint.get_pokemon_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "pokemon", data["results"])
        self._cache.load_documents(str(self.__class__.__name__), "location_area", data["results"])
        data = await self._client.request(Endpoint.get_pokemon_color_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "pokemon_color", data["results"])
        data = await self._client.request(Endpoint.get_pokemon_form_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "pokemon_form", data["results"])
        data = await self._client.request(Endpoint.get_pokemon_habitat_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "pokemon_habitat", data["results"])
        data = await self._client.request(Endpoint.get_pokemon_shape_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "pokemon_shape", data["results"])
        data = await self._client.request(Endpoint.get_pokemon_species_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "pokemon_species", data["results"])
        data = await self._client.request(Endpoint.get_stat_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "stat", data["results"])
        data = await self._client.request(Endpoint.get_type_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "type", data["results"])

    def get_ability(self, name: t.Union[str, int]) -> t.Optional[Ability]:
        """Get an ability by name or id.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the ability.

        Returns
        -------
        Optional[Ability]
            The ability if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the ability was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> ability = client.pokemon.get_ability("stench")
        >>> ability.id
        1
        """
        route = Endpoint.get_ability(name)
        self._validate_resource(self.cache.ability, name, route)
        return self.cache.ability.get(route, None)

    async def fetch_ability(self, name: t.Union[str, int]) -> t.Optional[Ability]:
        """Fetch an ability by name or id.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the ability.

        Returns
        -------
        Optional[Ability]
            The ability if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the ability was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     ability = await client.pokemon.fetch_ability("stench")
        ...     print(ability.id)
        ...     await client.close()
        >>> asyncio.run(main())
        1
        """
        route = Endpoint.get_ability(name)
        self._validate_resource(self.cache.ability, name, route)
        data = await self._client.request(route)
        return self.cache.ability.setdefault(route, Ability.from_payload(data))

    def get_characteristic(self, id_: int) -> t.Optional[Characteristic]:
        """Get a characteristic by id.

        Parameters
        ----------
        id_: int
            The id of the characteristic.

        Returns
        -------
        Optional[Characteristic]
            The characteristic if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the characteristic was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> characteristic = client.pokemon.get_characteristic(1)
        >>> characteristic.gene_modulo
        0
        """
        route = Endpoint.get_characteristic(id_)
        self._validate_resource(self.cache.characteristic, id_, route)
        return self.cache.characteristic.get(route, None)

    async def fetch_characteristic(self, id_: int) -> t.Optional[Characteristic]:
        """Fetch a characteristic by id.

        Parameters
        ----------
        id_: int
            The id of the characteristic.

        Returns
        -------
        Optional[Characteristic]
            The characteristic if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the characteristic was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     characteristic = await client.pokemon.fetch_characteristic(1)
        ...     print(characteristic.gene_modulo)
        ...     await client.close()
        >>> asyncio.run(main())
        0
        """
        route = Endpoint.get_characteristic(id_)
        self._validate_resource(self.cache.characteristic, id_, route)
        data = await self._client.request(route)
        return self.cache.characteristic.setdefault(route, Characteristic.from_payload(data))

    def get_egg_group(self, name: t.Union[str, int]) -> t.Optional[EggGroup]:
        """Get an egg group by name or id.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the egg group.

        Returns
        -------
        Optional[EggGroup]
            The egg group if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the egg group was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> egg_group = client.pokemon.get_egg_group("monster")
        >>> egg_group.id
        1
        """
        route = Endpoint.get_egg_group(name)
        self._validate_resource(self.cache.egg_group, name, route)
        return self.cache.egg_group.get(route, None)

    async def fetch_egg_group(self, name: t.Union[str, int]) -> t.Optional[EggGroup]:
        """Fetch an egg group by name or id.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the egg group.

        Returns
        -------
        Optional[EggGroup]
            The egg group if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the egg group was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     egg_group = await client.pokemon.fetch_egg_group("monster")
        ...     print(egg_group.id)
        ...     await client.close()
        >>> asyncio.run(main())
        1
        """
        route = Endpoint.get_egg_group(name)
        self._validate_resource(self.cache.egg_group, name, route)
        data = await self._client.request(route)
        return self.cache.egg_group.setdefault(route, EggGroup.from_payload(data))

    def get_gender(self, name: t.Union[str, int]) -> t.Optional[Gender]:
        """
        Get gender from cache.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the gender.

        Returns
        -------
        t.Optional[Gender]
            Gender model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if gender was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> gender = client.pokemon.get_gender(1)
        >>> gender.name
        'female'
        """
        route = Endpoint.get_gender(name)
        self._validate_resource(self.cache.gender, name, route)
        return self.cache.gender.get(route, None)

    async def fetch_gender(self, name: t.Union[str, int]) -> t.Optional[Gender]:
        """
        Fetches a gender model by name or id.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the

        Returns
        -------
        t.Optional[Gender]
            Gender model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if gender was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...    gender = await client.pokemon.fetch_gender(1)
        ...    print(gender.name)
        ...    await client.close()
        >>> asyncio.run(main())
        female
        """
        route = Endpoint.get_gender(name)
        self._validate_resource(self.cache.gender, name, route)
        data = await self._client.request(route)
        return self.cache.gender.setdefault(route, Gender.from_payload(data))

    def get_growth_rate(self, name: t.Union[str, int]) -> t.Optional[GrowthRate]:
        """
        Get growth rate from cache.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the growth rate.

        Returns
        -------
        t.Optional[GrowthRate]
            Growth rate model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if growth rate was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> growth_rate = client.pokemon.get_growth_rate(1)
        >>> growth_rate.name
        'slow'
        """
        route = Endpoint.get_growth_rate(name)
        self._validate_resource(self.cache.growth_rate, name, route)
        return self.cache.growth_rate.get(route, None)

    async def fetch_growth_rate(self, name: t.Union[str, int]) -> t.Optional[GrowthRate]:
        """
        Fetches a growth rate model by name or id.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the

        Returns
        -------
        t.Optional[GrowthRate]
            Growth rate model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if growth rate was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...    growth_rate = await client.pokemon.fetch_growth_rate(1)
        ...    print(growth_rate.name)
        ...    await client.close()
        >>> asyncio.run(main())
        slow
        """
        route = Endpoint.get_growth_rate(name)
        self._validate_resource(self.cache.growth_rate, name, route)
        data = await self._client.request(route)
        return self.cache.growth_rate.setdefault(route, GrowthRate.from_payload(data))

    def get_nature(self, name: t.Union[str, int]) -> t.Optional[Nature]:
        """
        Get nature from cache.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the nature.

        Returns
        -------
        t.Optional[Nature]
            Nature model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if nature was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> nature = client.pokemon.get_nature(1)
        >>> nature.name
        'hardy'
        """
        route = Endpoint.get_nature(name)
        self._validate_resource(self.cache.nature, name, route)
        return self.cache.nature.get(route, None)

    async def fetch_nature(self, name: t.Union[str, int]) -> t.Optional[Nature]:
        """
        Fetches a nature model by name or id.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the

        Returns
        -------
        t.Optional[Nature]
            Nature model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if nature was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...    nature = await client.pokemon.fetch_nature(1)
        ...    print(nature.name)
        ...    await client.close()
        >>> asyncio.run(main())
        hardy
        """
        route = Endpoint.get_nature(name)
        self._validate_resource(self.cache.nature, name, route)
        data = await self._client.request(route)
        return self.cache.nature.setdefault(route, Nature.from_payload(data))

    def get_pokeathlon_stat(self, name: t.Union[str, int]) -> t.Optional[PokeathlonStat]:
        """
        Get pokeathlon stat from cache.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the pokeathlon stat.

        Returns
        -------
        t.Optional[PokeathlonStat]
            Pokeathlon stat model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if pokeathlon stat was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> pokeathlon_stat = client.pokemon.get_pokeathlon_stat(1)
        >>> pokeathlon_stat.name
        'speed'
        """
        route = Endpoint.get_pokeathlon_stat(name)
        self._validate_resource(self.cache.pokeathlon_stat, name, route)
        return self.cache.pokeathlon_stat.get(route, None)

    async def fetch_pokeathlon_stat(self, name: t.Union[str, int]) -> t.Optional[PokeathlonStat]:
        """
        Fetches a pokeathlon stat model by name or id.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the

        Returns
        -------
        t.Optional[PokeathlonStat]
            Pokeathlon stat model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if pokeathlon stat was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...    pokeathlon_stat = await client.pokemon.fetch_pokeathlon_stat(1)
        ...    print(pokeathlon_stat.name)
        ...    await client.close()
        >>> asyncio.run(main())
        speed
        """
        route = Endpoint.get_pokeathlon_stat(name)
        self._validate_resource(self.cache.pokeathlon_stat, name, route)
        data = await self._client.request(route)
        return self.cache.pokeathlon_stat.setdefault(route, PokeathlonStat.from_payload(data))

    def get_location_area_encounter(self, name: t.Union[str, int]) -> t.Optional[LocationAreaEncounter]:
        """
        Get location area encounter from cache.
        It gets areas where a pokemon can be found.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the location area encounter.

        Returns
        -------
        t.Optional[LocationAreaEncounter]
            Location area encounter model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if location area encounter was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> location_area_encounter = client.pokemon.get_location_area_encounter(1)
        >>> location_area_encounter.location_area.name
        'cerulean-city-area'
        """
        route = Endpoint.get_location_area_encounter(name)
        self._validate_resource(self.cache.location_area, name, route)
        return self.cache.location_area.get(route, None)

    async def fetch_location_area_encounter(self, name: t.Union[str, int]) -> t.Optional[LocationAreaEncounter]:
        """
        Fetches a location area encounter model by name or id.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the

        Returns
        -------
        t.Optional[LocationAreaEncounter]
            Location area encounter model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if location area encounter was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...    location_area_encounter = await client.pokemon.fetch_location_area_encounter(1)
        ...    print(location_area_encounter.location_area.name)
        ...    await client.close()
        >>> asyncio.run(main())
        cerulean-city-area
        """
        route = Endpoint.get_location_area_encounter(name)
        self._validate_resource(self.cache.location_area, name, route)
        data = await self._client.request(route)
        return self.cache.location_area.setdefault(route, LocationAreaEncounter.from_payload(data[0]))

    def get_pokemon(self, name: t.Union[str, int]) -> t.Optional[PokemonModel]:
        """
        Get pokemon from cache.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the pokemon.

        Returns
        -------
        t.Optional[PokemonModel]
            Pokemon model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if pokemon was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> pokemon = client.pokemon.get_pokemon(1)
        >>> pokemon.name
        'bulbasaur'
        """
        route = Endpoint.get_pokemon(name)
        self._validate_resource(self.cache.pokemon, name, route)
        return self.cache.pokemon.get(route, None)

    async def fetch_pokemon(self, name: t.Union[str, int]) -> t.Optional[PokemonModel]:
        """
        Fetches a pokemon model by name or id.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the

        Returns
        -------
        t.Optional[PokemonModel]
            Pokemon model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if pokemon was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...    pokemon = await client.pokemon.fetch_pokemon(1)
        ...    print(pokemon.name)
        ...    await client.close()
        >>> asyncio.run(main())
        bulbasaur
        """
        route = Endpoint.get_pokemon(name)
        self._validate_resource(self.cache.pokemon, name, route)
        data = await self._client.request(route)
        return self.cache.pokemon.setdefault(route, PokemonModel.from_payload(data))

    def get_pokemon_color(self, name: t.Union[str, int]) -> t.Optional[PokemonColor]:
        """
        Get pokemon color from cache.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the pokemon color.

        Returns
        -------
        t.Optional[PokemonColor]
            Pokemon color model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if pokemon color was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> pokemon_color = client.pokemon.get_pokemon_color(1)
        >>> pokemon_color.name
        'black'
        """
        route = Endpoint.get_pokemon_color(name)
        self._validate_resource(self.cache.pokemon_color, name, route)
        return self.cache.pokemon_color.get(route, None)

    async def fetch_pokemon_color(self, name: t.Union[str, int]) -> t.Optional[PokemonColor]:
        """
        Fetches a pokemon color model by name or id.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the

        Returns
        -------
        t.Optional[PokemonColor]
            Pokemon color model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if pokemon color was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...    pokemon_color = await client.pokemon.fetch_pokemon_color(1)
        ...    print(pokemon_color.name)
        ...    await client.close()
        >>> asyncio.run(main())
        black
        """
        route = Endpoint.get_pokemon_color(name)
        self._validate_resource(self.cache.pokemon_color, name, route)
        data = await self._client.request(route)
        return self.cache.pokemon_color.setdefault(route, PokemonColor.from_payload(data))

    def get_pokemon_form(self, name: t.Union[str, int]) -> t.Optional[PokemonForm]:
        """
        Get pokemon form from cache.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the pokemon form.

        Returns
        -------
        t.Optional[PokemonForm]
            Pokemon form model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if pokemon form was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> pokemon_form = client.pokemon.get_pokemon_form(1)
        >>> pokemon_form.name
        'bulbasaur'
        """
        route = Endpoint.get_pokemon_form(name)
        self._validate_resource(self.cache.pokemon_form, name, route)
        return self.cache.pokemon_form.get(route, None)

    async def fetch_pokemon_form(self, name: t.Union[str, int]) -> t.Optional[PokemonForm]:
        """
        Fetches a pokemon form model by name or id.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the

        Returns
        -------
        t.Optional[PokemonForm]
            Pokemon form model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if pokemon form was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...    pokemon_form = await client.pokemon.fetch_pokemon_form(1)
        ...    print(pokemon_form.name)
        ...    await client.close()
        >>> asyncio.run(main())
        bulbasaur
        """
        route = Endpoint.get_pokemon_form(name)
        self._validate_resource(self.cache.pokemon_form, name, route)
        data = await self._client.request(route)
        return self.cache.pokemon_form.setdefault(route, PokemonForm.from_payload(data))

    def get_pokemon_habitat(self, name: t.Union[str, int]) -> t.Optional[PokemonHabitats]:
        """
        Get pokemon habitat from cache.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the pokemon habitat.

        Returns
        -------
        t.Optional[PokemonHabitats]
            Pokemon habitat model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if pokemon habitat was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> pokemon_habitat = client.pokemon.get_pokemon_habitat(1)
        >>> pokemon_habitat.name
        'cave'
        """
        route = Endpoint.get_pokemon_habitat(name)
        self._validate_resource(self.cache.pokemon_habitat, name, route)
        return self.cache.pokemon_habitat.get(route, None)

    async def fetch_pokemon_habitat(self, name: t.Union[str, int]) -> t.Optional[PokemonHabitats]:
        """
        Fetches a pokemon habitat model by name or id.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the

        Returns
        -------
        t.Optional[PokemonHabitats]
            Pokemon habitat model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if pokemon habitat was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...    pokemon_habitat = await client.pokemon.fetch_pokemon_habitat(1)
        ...    print(pokemon_habitat.name)
        ...    await client.close()
        >>> asyncio.run(main())
        cave
        """
        route = Endpoint.get_pokemon_habitat(name)
        self._validate_resource(self.cache.pokemon_habitat, name, route)
        data = await self._client.request(route)
        return self.cache.pokemon_habitat.setdefault(route, PokemonHabitats.from_payload(data))

    def get_pokemon_shape(self, name: t.Union[str, int]) -> t.Optional[PokemonShape]:
        """
        Get pokemon shape from cache.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the pokemon shape.

        Returns
        -------
        t.Optional[PokemonShape]
            Pokemon shape model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if pokemon shape was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> pokemon_shape = client.pokemon.get_pokemon_shape(1)
        >>> pokemon_shape.name
        'ball'
        """
        route = Endpoint.get_pokemon_shape(name)
        self._validate_resource(self.cache.pokemon_shape, name, route)
        return self.cache.pokemon_shape.get(route, None)

    async def fetch_pokemon_shape(self, name: t.Union[str, int]) -> t.Optional[PokemonShape]:
        """
        Fetches a pokemon shape model by name or id.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the

        Returns
        -------
        t.Optional[PokemonShape]
            Pokemon shape model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if pokemon shape was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...    pokemon_shape = await client.pokemon.fetch_pokemon_shape(1)
        ...    print(pokemon_shape.name)
        ...    await client.close()
        >>> asyncio.run(main())
        ball
        """
        route = Endpoint.get_pokemon_shape(name)
        self._validate_resource(self.cache.pokemon_shape, name, route)
        data = await self._client.request(route)
        return self.cache.pokemon_shape.setdefault(route, PokemonShape.from_payload(data))

    def get_pokemon_species(self, name: t.Union[str, int]) -> t.Optional[PokemonSpecies]:
        """
        Get pokemon species from cache.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the pokemon species.

        Returns
        -------
        t.Optional[PokemonSpecies]
            Pokemon species model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if pokemon species was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> pokemon_species = client.pokemon.get_pokemon_species(1)
        >>> pokemon_species.name
        'bulbasaur'
        """
        route = Endpoint.get_pokemon_species(name)
        self._validate_resource(self.cache.pokemon_species, name, route)
        return self.cache.pokemon_species.get(route, None)

    async def fetch_pokemon_species(self, name: t.Union[str, int]) -> t.Optional[PokemonSpecies]:
        """
        Fetches a pokemon species model by name or id.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the

        Returns
        -------
        t.Optional[PokemonSpecies]
            Pokemon species model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if pokemon species was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...    pokemon_species = await client.pokemon.fetch_pokemon_species(1)
        ...    print(pokemon_species.name)
        ...    await client.close()
        >>> asyncio.run(main())
        bulbasaur
        """
        route = Endpoint.get_pokemon_species(name)
        self._validate_resource(self.cache.pokemon_species, name, route)
        data = await self._client.request(route)
        return self.cache.pokemon_species.setdefault(route, PokemonSpecies.from_payload(data))

    def get_stat(self, name: t.Union[str, int]) -> t.Optional[Stat]:
        """
        Get stat from cache.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the stat.

        Returns
        -------
        t.Optional[Stat]
            Stat model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if stat was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> stat = client.pokemon.get_stat(1)
        >>> stat.name
        'hp'
        """
        route = Endpoint.get_stat(name)
        self._validate_resource(self.cache.stat, name, route)
        return self.cache.stat.get(route, None)

    async def fetch_stat(self, name: t.Union[str, int]) -> t.Optional[Stat]:
        """
        Fetches a stat model by name or id.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the

        Returns
        -------
        t.Optional[Stat]
            Stat model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if stat was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...    stat = await client.pokemon.fetch_stat(1)
        ...    print(stat.name)
        ...    await client.close()
        >>> asyncio.run(main())
        hp
        """
        route = Endpoint.get_stat(name)
        self._validate_resource(self.cache.stat, name, route)
        data = await self._client.request(route)
        return self.cache.stat.setdefault(route, Stat.from_payload(data))

    def get_type(self, name: t.Union[str, int]) -> t.Optional[Type]:
        """
        Get type from cache.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the type.

        Returns
        -------
        t.Optional[Type]
            Type model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if type was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> type_ = client.pokemon.get_type(1)
        >>> type_.name
        'normal'
        """
        route = Endpoint.get_type(name)
        self._validate_resource(self.cache.type, name, route)
        return self.cache.type.get(route, None)

    async def fetch_type(self, name: t.Union[str, int]) -> t.Optional[Type]:
        """
        Fetches a type model by name or id.

        Parameters
        ----------
        name: Union[str, int]
            The name or id of the

        Returns
        -------
        t.Optional[Type]
            Type model if found, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            if type was not found.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...    type_ = await client.pokemon.fetch_type(1)
        ...    print(type_.name)
        ...    await client.close()
        >>> asyncio.run(main())
        normal
        """
        route = Endpoint.get_type(name)
        self._validate_resource(self.cache.type, name, route)
        data = await self._client.request(route)
        return self.cache.type.setdefault(route, Type.from_payload(data))


def setup(lance: "PokeLance") -> None:
    """Setup the pokemon cog."""
    lance.add_extension("pokemon", Pokemon(lance.http))



import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import Description, MachineVersionDetail, Name, NamedResource, Resource, VerboseEffect

from .utils import (
    AbilityEffectChange,
    ContestComboSet,
    MoveFlavorText,
    MoveMetaData,
    MoveStatChange,
    PastMoveStatValues,
)

__all__: t.Tuple[str, ...] = (
    "Move",
    "MoveAilment",
    "MoveBattleStyle",
    "MoveCategory",
    "MoveDamageClass",
    "MoveLearnMethod",
    "MoveTarget",
)


@attrs.define(slots=True, kw_only=True)
class Move(BaseModel):
    """Move model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    accuracy: int
        The percent value of how likely this move is to be successful.
    effect_chance: int
        The percent value of how likely it is this moves effect will take effect.
    pp: int
        Power points. The number of times this move can be used.
    priority: int
        A value between -8 and 8. Sets the order in which moves
        are executed during battle. See Bulbapedia for greater detail.
    power: int
        The base power of this move with a value of 0 if it does not have a base power.
    contest_combos: ContestComboSet
        A detail of normal and super contest combos that require this move.
    contest_type: NamedResource
        The type of appeal this move gives a Pokémon when used in a contest.
    contest_effect: Resource
        The effect the move has when used in a contest.
    damage_class: NamedResource
        The type of damage the move inflicts on the target, e.g. physical.
    effect_entries: t.List[VerboseEffect]
        The effect of this move listed in different languages.
    effect_changes: t.List[AbilityEffectChange]
        The list of previous effects this move has had across version groups of the games.
    flavor_text_entries: t.List[MoveFlavorText]
        The flavor text of this move listed in different languages.
    generation: NamedResource
        The generation in which this move was introduced.
    machines: t.List[MachineVersionDetail]
        A list of the machines that teach this move.
    meta: MoveMetaData
        Metadata about this move.
    names: t.List[Name]
        The name of this resource listed in different languages.
    past_values: t.List[PastMoveStatValues]
        A list of move resource value changes across version groups of the games.
    stat_changes: t.List[MoveStatChange]
        A list of stats this moves effects and how much it effects them.
    super_contest_effect: Resource
        The effect the move has when used in a super contest.
    target: NamedResource
        The type of target that will receive the effects of the attack.
    type: NamedResource
        The elemental type of this move.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    accuracy: int = attrs.field(factory=int)
    effect_chance: int = attrs.field(factory=int)
    pp: int = attrs.field(factory=int)
    priority: int = attrs.field(factory=int)
    power: int = attrs.field(factory=int)
    contest_combos: ContestComboSet = attrs.field(factory=ContestComboSet)
    contest_type: NamedResource = attrs.field(factory=NamedResource)
    contest_effect: Resource = attrs.field(factory=Resource)
    damage_class: NamedResource = attrs.field(factory=NamedResource)
    effect_entries: t.List[VerboseEffect] = attrs.field(factory=list)
    effect_changes: t.List[AbilityEffectChange] = attrs.field(factory=list)
    learned_by_pokemon: t.List[NamedResource] = attrs.field(factory=list)
    flavor_text_entries: t.List[MoveFlavorText] = attrs.field(factory=list)
    generation: NamedResource = attrs.field(factory=NamedResource)
    machines: t.List[MachineVersionDetail] = attrs.field(factory=list)
    meta: MoveMetaData = attrs.field(factory=MoveMetaData)
    names: t.List[Name] = attrs.field(factory=list)
    past_values: t.List[PastMoveStatValues] = attrs.field(factory=list)
    stat_changes: t.List[MoveStatChange] = attrs.field(factory=list)
    super_contest_effect: Resource = attrs.field(factory=Resource)
    target: NamedResource = attrs.field(factory=NamedResource)
    type: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Move":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            accuracy=payload.get("accuracy", 0),
            effect_chance=payload.get("effect_chance", 0),
            pp=payload.get("pp", 0),
            priority=payload.get("priority", 0),
            power=payload.get("power", 0),
            contest_combos=ContestComboSet.from_payload(payload.get("contest_combos", {})),
            contest_type=NamedResource.from_payload(payload.get("contest_type", {}) or {}),
            contest_effect=Resource.from_payload(payload.get("contest_effect", {}) or {}),
            damage_class=NamedResource.from_payload(payload.get("damage_class", {}) or {}),
            effect_entries=[VerboseEffect.from_payload(i) for i in payload.get("effect_entries", [])],
            effect_changes=[AbilityEffectChange.from_payload(i) for i in payload.get("effect_changes", [])],
            learned_by_pokemon=[NamedResource.from_payload(i) for i in payload.get("learned_by_pokemon", [])],
            flavor_text_entries=[MoveFlavorText.from_payload(i) for i in payload.get("flavor_text_entries", [])],
            generation=NamedResource.from_payload(payload.get("generation", {}) or {}),
            machines=[MachineVersionDetail.from_payload(i) for i in payload.get("machines", [])],
            meta=MoveMetaData.from_payload(payload.get("meta", {})),
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            past_values=[PastMoveStatValues.from_payload(i) for i in payload.get("past_values", [])],
            stat_changes=[MoveStatChange.from_payload(i) for i in payload.get("stat_changes", [])],
            super_contest_effect=Resource.from_payload(payload.get("super_contest_effect", {}) or {}),
            target=NamedResource.from_payload(payload.get("target", {}) or {}),
            type=NamedResource.from_payload(payload.get("type", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class MoveAilment(BaseModel):
    """
    MoveAilment models the data returned by the API for a move ailment.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    moves: t.List[NamedResource]
        A list of moves that cause this ailment.
    names: t.List[Name]
        The name of this resource listed in different languages.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    moves: t.List[NamedResource] = attrs.field(factory=list)
    names: t.List[Name] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "MoveAilment":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            moves=[NamedResource.from_payload(i) for i in payload.get("moves", [])],
            names=[Name.from_payload(i) for i in payload.get("names", [])],
        )


@attrs.define(slots=True, kw_only=True)
class MoveBattleStyle(BaseModel):
    """
    MoveBattleStyle models the data returned by the API for a move battle style.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    names: t.List[Name]
        The name of this resource listed in different languages.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    names: t.List[Name] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "MoveBattleStyle":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            names=[Name.from_payload(i) for i in payload.get("names", [])],
        )


@attrs.define(slots=True, kw_only=True)
class MoveCategory(BaseModel):
    """
    MoveCategory models the data returned by the API for a move category.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    moves: t.List[NamedResource]
        A list of moves that fall into this category.
    descriptions: t.List[Description]
        The description of this resource listed in different languages.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    moves: t.List[NamedResource] = attrs.field(factory=list)
    descriptions: t.List[Description] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "MoveCategory":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            moves=[NamedResource.from_payload(i) for i in payload.get("moves", [])],
            descriptions=[Description.from_payload(i) for i in payload.get("descriptions", [])],
        )


@attrs.define(slots=True, kw_only=True)
class MoveDamageClass(BaseModel):
    """
    MoveDamageClass models the data returned by the API for a move damage class.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    descriptions: t.List[Description]
        The description of this resource listed in different languages.
    moves: t.List[NamedResource]
        A list of moves that fall into this damage class.
    names: t.List[Name]
        The name of this resource listed in different languages.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    descriptions: t.List[Description] = attrs.field(factory=list)
    moves: t.List[NamedResource] = attrs.field(factory=list)
    names: t.List[Name] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "MoveDamageClass":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            descriptions=[Description.from_payload(i) for i in payload.get("descriptions", [])],
            moves=[NamedResource.from_payload(i) for i in payload.get("moves", [])],
            names=[Name.from_payload(i) for i in payload.get("names", [])],
        )


@attrs.define(slots=True, kw_only=True)
class MoveLearnMethod(BaseModel):
    """
    MoveLearnMethod models the data returned by the API for a move learn method.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    descriptions: t.List[Description]
        The description of this resource listed in different languages.
    names: t.List[Name]
        The name of this resource listed in different languages.
    version_groups: t.List[NamedResource]
        A list of version groups where moves can be learned through this method.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    descriptions: t.List[Description] = attrs.field(factory=list)
    names: t.List[Name] = attrs.field(factory=list)
    version_groups: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "MoveLearnMethod":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            descriptions=[Description.from_payload(i) for i in payload.get("descriptions", [])],
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            version_groups=[NamedResource.from_payload(i) for i in payload.get("version_groups", [])],
        )


@attrs.define(slots=True, kw_only=True)
class MoveTarget(BaseModel):
    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    descriptions: t.List[Description] = attrs.field(factory=list)
    moves: t.List[NamedResource] = attrs.field(factory=list)
    names: t.List[Name] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "MoveTarget":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            descriptions=[Description.from_payload(i) for i in payload.get("descriptions", [])],
            moves=[NamedResource.from_payload(i) for i in payload.get("moves", [])],
            names=[Name.from_payload(i) for i in payload.get("names", [])],
        )



import typing as t

from .berry import Berry, BerryFirmness, BerryFlavor
from .contest import ContestEffect, ContestType, SuperContestEffect
from .encounter import EncounterCondition, EncounterConditionValue, EncounterMethod
from .evolution import EvolutionChain, EvolutionTrigger
from .game import Generation, Pokedex, Version, VersionGroup
from .item import Item, ItemAttribute, ItemCategory, ItemFlingEffect, ItemPocket
from .location import Location, LocationArea, PalParkArea, Region
from .machine import Machine
from .move import Move, MoveAilment, MoveBattleStyle, MoveCategory, MoveDamageClass, MoveLearnMethod, MoveTarget
from .pokemon import (
    Ability,
    Characteristic,
    EggGroup,
    Gender,
    GrowthRate,
    LocationAreaEncounter,
    Nature,
    PokeathlonStat,
    Pokemon,
    PokemonColor,
    PokemonForm,
    PokemonHabitats,
    PokemonShape,
    PokemonSpecies,
    Stat,
    Type,
)

__all__: t.Tuple[str, ...] = (
    "Berry",
    "BerryFirmness",
    "BerryFlavor",
    "ContestEffect",
    "ContestType",
    "SuperContestEffect",
    "EncounterCondition",
    "EncounterConditionValue",
    "EncounterMethod",
    "EvolutionChain",
    "EvolutionTrigger",
    "Generation",
    "Pokedex",
    "Version",
    "VersionGroup",
    "Item",
    "ItemAttribute",
    "ItemCategory",
    "ItemFlingEffect",
    "ItemPocket",
    "Location",
    "LocationArea",
    "PalParkArea",
    "Region",
    "Machine",
    "Move",
    "MoveAilment",
    "MoveBattleStyle",
    "MoveCategory",
    "MoveDamageClass",
    "MoveLearnMethod",
    "MoveTarget",
    "Ability",
    "Characteristic",
    "Pokemon",
    "PokemonSpecies",
    "PokemonForm",
    "Nature",
    "Type",
    "Gender",
    "GrowthRate",
    "EggGroup",
    "LocationAreaEncounter",
    "PokemonColor",
    "PokeathlonStat",
    "PokemonHabitats",
    "PokemonShape",
    "Stat",
)
